#!/usr/bin/python
## \file agar.py

## \brief The main Agar object.  Also the file to execute
## for performing regression and coverage tests.
##

import os
import sys
import string
import gzip
import time

from xml.sax.saxutils import escape, quoteattr

import submissions
import tests
import rubric
import comments
import gnumeripy
import loader

class Agar:
    """\brief The main Agar object.
    
    This is the main functionality class.  Upon initialization it
    attempts to find submissions located in the current directory, or
    in the optional 'basedir' parameter.  It also discovers the valid
    Agar tests stored in the tests/ directory off the main Agar
    directory.

    Major Members:
    
    * self.submissions: a Submissions object (read-only dictionary-like
    container) containing all of the current Submission objects
    
    * self.tests: a Tests object (read-only dictionary-like container)
      containing all of the Test objects

    * self.rubric: a Rubric object defining the tree-hierarchy of
      RubricTests (tests with associated configuration)

    * self.userComments: A dictionary mapping names to user-created
      comments.  Alterations to this should be mediated with the
      add/del methods.

    Minor Members:

    * self.basedir: the directory storing all of the submissions

    * self.agarPath: the directory containing agar.py
    
    * self.config: a dictionary mapping to site-specific configuration

    """
    def __init__(self, basedir=None, filemask=None):
        """
        Construct a new Agar object.

        Optional Parameters:
        
        @param basedir (\c string).  Use this directory as the parent of all
        submission directories.  If not specified, use os.getcwd().
        """
        if basedir == None:
            basedir = os.getcwd()

        assert(type(basedir) == type(""))

        self.basedir = basedir ## The directory that contains all submissions
        self.config = self.readConfig() ## Site-specific configuration

        # The name of this assignment
        assignmentName = self.basedir
        while assignmentName[-1] == "/":
            assignmentName = assignmentName[:-1]
        assignmentName = assignmentName.split("/")[-1]
        self.setAssignmentName(assignmentName)

        ## The main submissions object
        self.submissions = submissions.Submissions(self.config,
                                                   self.basedir,
                                                   filemask)

        # Get the tests

        ## The path to the agar executable
        self.agarPath = os.path.dirname(os.path.abspath(sys.argv[0]))

        ## The main tests object
        self.tests = tests.Tests(self.agarPath + "/tests/")

        ## The main rubric object
        self.rubric = rubric.Rubric(self)

        ## The user comments dictionary, maps strings to comments
        self.userComments = comments.UserComments()

        # Mail info
        self.reply = ""
        self.cclist = ""
        self.domain = "cs.ucr.edu"

    def getAssignmentName(self):
        return self.assignmentName

    def setAssignmentName(self, val):
        goodChars = " ." + string.lowercase + string.uppercase + string.digits
        for char in val:
            assert(char in goodChars)
        self.assignmentName = val
    
    def readConfig(self, configFile=None):
        """
        Read in site-specific configuration.  Generally this will be
        things like existance of and path to optional binaries, and
        directory structure for submissions.
        
        @param configFile -- a string path to the desired configuration file
        (currently ignored)

        TODO: Find an elegant way of doing this.
        """
        
        if configFile == None:
            configFile = os.path.expanduser("~/.agar2cfg")

        assert(type(configFile) == type(""))

        # Our return value
        ret = {}
        
        # Set defaults
        ret["isSubmissionDir"] = defaultIsSubDir
        ret["isOntime"] = defaultIsOntime
        ret["getFiles"] = defaultGetFiles

        return ret

    def exportCSVScores(self, filename):
        """Generate a simple and universally-readable summary of scores.
        
        @param filename -- name of a (writable) file to export CVS
        formatted results into.
        """
        scores = [self.submissions[subname].getScores(self.rubric, name) \
                   for subname in self.submissions \
                   for name in self.submissions[subname].getUsernames()]
        scores.sort()

        headers = [["username", "Total"] + self.rubric.getNames()]

        scores = headers + scores

        of = open(filename, 'w')
        for row in scores:
            s = ""
            for val in row:
                s += str(val) + ","
            print >>of, s[:-1]

    def reorderRows(self, sheet, scores):
        header = sheet.data[0][:]
        for ind, val in enumerate(header):
            if val.lower() in ["username", "login"]:
                sheet.setIndexCol(ind)
                break

        ret = [scores.pop(0)]
        unknowns = []
        for row in scores:
            try:
                ind = sheet.getKeyRow(row[0])
                while len(ret) <= ind:
                    ret.append([])
                ret[ind] = row
            except:
                unknowns.append(row)

        while len(ret) <= sheet.rows():
            ret.append([])

        ret += unknowns
        return ret
                
        
        
    def exportGnumericScores(self,
                             filename,
                             summary=False,
                             overwrite=True,
                             merge=False,
                             g=None):
        """Export scores into a gnumeric spreadsheet.
        
        @param filename -- name of an existing .gnumeric file to export
        scores into.
        @param summary -- Do we want to add a column to the summary sheet?
        @param overwrite -- Do we overwrite existing data?
        @param merge -- Do we merge into existing sheets or just overwrite 'em?
        """

        scores = [ self.submissions[subname].getScores(self.rubric, name) \
                   for subname in self.submissions \
                   for name in self.submissions[subname].getUsernames()]
        scores.sort()

        headers = [["username", "Total"] + self.rubric.getNames()]

        if g is None:
            g = gnumeripy.GnumeriPy(filename)
            
        scores = headers + scores
        scores = self.reorderRows(g["Summary"], scores)

        # Do the summary export
        if summary:
            summary = g["Summary"]
            try:
                colInd = summary.getKeyCol(self.assignmentName)
            except ValueError:
                colInd = summary.cols()
                summary.set(0, colInd, self.assignmentName)

            for ind, row in enumerate(scores):
                if row and ind > 0:
                    summary.set(ind, colInd, \
                            ("=%s!B%d" % (self.assignmentName, ind + 1), None))

        if self.assignmentName not in g.sheetNames():
            g.addSheet(self.assignmentName)
            g[self.assignmentName].data = scores
            g.saveAs(filename)
            return

        # Now, we know there is a sheet of this name already, what do
        # we do with it?

        # It already exists and we were told not to meddle.
        if not merge and not overwrite:
            raise UserError
            
        # Overwrite the sheet, don't merge entries
        if not merge and overwrite:
            g[self.assignmentName].data = scores
            g.saveAs(filename)
            return

        s = g[self.assignmentName]

        # Merge into the sheet, keep old
        if merge and not overwrite:
            for ind, row in enumerate(scores):
                if not s.data[ind]:
                    s.data[ind] = row

        elif merge and overwrite:
            for ind, row in enumerate(scores):
                if row:
                    s.data[ind] = row
            
        g.saveAs(filename)
        return

    def saveAs(self, filename, subs=True, rubric=True, comments=True):
        print >>gzip.open(filename, 'w'), \
              "\n".join(self.toXML(subs, rubric, comments))

    def toXML(self, subs, rubric, comments):
        ret = []
        ret.append('<?xml version="1.0" encoding="UTF-8"?>')
        ret.append('<agar xmlns="http://www.cs.ucr.edu/~titus/agar.dtd">')
        ret.append('<version epoch="0" major="5" minor="0" full="0.5.0" />')
        ret.append('<mail reply=%s cclist=%s domain=%s />' % \
                   (quoteattr(self.reply),
                    quoteattr(self.cclist),
                    quoteattr(self.domain)))
        ret.append('<project name=%s basedir=%s />' % \
                   (quoteattr(self.assignmentName),
                    quoteattr(os.path.realpath(self.submissions.basedir))))
        if subs:
            ret += self.submissions.toXML()
        if rubric:
            ret += self.rubric.toXML()
        if comments:
            ret += self.userComments.toXML()
        ret.append("</agar>")

        depth = 0
        for ind, val in enumerate(ret):
            if not val: continue

            # Escape rogue quotes in char data
            if val[0] != "<":
                ret[ind] = val.replace("'", "\'")
                continue

            # Calculate spacing
            spaces = " " * depth

            # Figure out our indentation level
            if "/>" not in val and "</" not in val:
                depth += 1
            if val.strip().startswith("</"):
                depth -= 1
                spaces = " " * depth

            # Add in indentation
            val = spaces + val
                
            ret[ind] = val
        
        return ret

    def load(self, filename):
        xml = gzip.open(filename).readlines()
        self.parseXML(xml)

    def parseXML(self, xml):
        (version, subs, cs, rubricData, order,
         reply, cclist, domain,
         name, basedir) = \
                  loader.Load(xml)

        self.reply = reply
        self.cclist = cclist
        self.domain = domain
        self.assignmentName = name

        self.submissions.basedir = basedir
        os.chdir(os.path.dirname(basedir))

        if rubricData:
            self.rubric.set(rubricData, order)
        if subs:
            self.submissions.set(subs, self.rubric)            
        if cs:
            self.userComments.set(cs, self.rubric)        

    def merge(self, fn, r, c, s):
        # 0 = old only
        # 1 = merge, prefer old
        # 2 = merge, prefer new
        # 3 = new only

        xml = gzip.open(fn).readlines()
        (version, subs, cs, rubricData, order) = loader.Load(xml)[:5]

        if r == 3:
            self.rubric.set(rubricData, order)
        elif r == 2:
            self.rubric.merge(rubricData, order, False)
        elif r == 1:
            self.rubric.merge(rubricData, order, True)

        if c == 3:
            self.userComments.set(cs)
        elif c == 2:
            self.userComments.merge(cs, False, self.submissions)
        elif c == 1:
            self.userComments.merge(cs, True, self.submissions)

        self.submissions.merge(subs)    

        # Now we need to go through everything and make sure tests are
        # pointing to the right thing
        for cname in self.userComments:
            c = self.userComments[cname]
            c.test = self.rubric[c.test.name]

        for sname in self.submissions:
            sub = self.submissions[sname]
            for c in sub.comments:
                c.test = self.rubric[c.test.name]                           
        
    def updatePoints(self, test, val):
        """Update the max-points value for test to be val.  This will
        call updateTestScores on every submission.

        @param test -- instance of Test
        @param val -- number
        """
        assert(isinstance(test, rubric.RubricTest))
        assert(type(val) in [type(0), type(0.0)])
        
        ratio = val / float(test.getPoints())
        test.setPoints(val)

        for sub in self.submissions:
            self.submissions[sub].updateTestScores(test, ratio)

    def convertForAnnotation(self, callback=None):
        maxN = len(self.submissions)
        n = 0

        for name in self.submissions:
            sub = self.submissions[name]

            if callback:
                callback(name, n, maxN)

            self.submissions[name].convertForAnnotation(self.agarPath)
            n += 1

        if callback:
            callback(name, n, maxN)               

    def addUserComment(self, name, test, text, op, value, appliesToAll=False):
        """Put together a new user comment.

        @param name The name that will be mapped to this comment
        @param test A toplevel rubric test for this comment
        @param text The feedback text of this comment
        @param op Either '=', '+', or '*'
        @param value The value (float) for this comment
        @param appliesToAll A bool, does this comment apply to all top levels?
        """
        assert(name not in self.userComments)
        assert(isinstance(test, rubric.RubricTest))
        assert(test in self.rubric.getTopLevel())
        assert(type(text) == type(""))
        assert(op in ["+", "=", "*"])
        assert(type(value) in [type(0), type(0.0)])

        newComment = comments.Comment(test, False, value,
                                      text, op, appliesToAll)
        self.userComments[name] = newComment

    def delUserComment(self, name):
        """If there is a user-defined (that is, non-automatic) comment
        with this name, then remove it from all submissions and remove
        it from the master list.

        Calls removeAllCopiesOfComment

        @param name The name of the comment to remove"""
        assert(name in self.userComments)

        self.removeAssignedComment(self.userComments[name])
        del self.userComments[name]

    def removeAllCopiesOfComment(self, comment):
        """Remove all assigned copies of the given comment

        @param comment The comment to remove from submissions"""
        assert(isinstance(comment, comments.Comment))
        # Remove from all submissions to which this has been assigned
        for username in self.submissions:
            self.submissions[username].removeComment(comment)

    def assignUserComment(self, commentName, subName):
        """Look up the comment with the given name, and assign it to
        the submission with the given name.

        @param commentName The name of this comment
        @param subName The name of the submission
        """
        assert(commentName in self.userComments)
        assert(subName in self.submissions)
        
        self.submissions[subName].comments.addComment(
            self.userComments[commentName])

    def removeUserComment(self, commentName, subName):
        """Unassign the comment with the given name from the
        submission with the given name.

        @param commentName The name of the comment to remove
        @param subName The name of the submission to remove it from""" 
        assert(commentName in self.userComments)
        assert(subName in self.submissions)

        self.submissions[subName].comments.removeComment(
            self.userComments[commentName])

    def clearAllComments(self):
        """Remove all assigned comments, automatically generated or
        otherwise.""" 
        for subname in self.submissions:
            self.submissions[subname].comments.clear()

    def clearAllExecutables(self):
        """Unset the executable for all submissions."""
        for subname in self.submissions:
            self.submissions[subname].executable = None

    def runTest(self, test, callback=None):
        """Run all submissions through this test.

        @param test The rubric test to run
        """
        assert(isinstance(test, rubric.RubricTest))
        numTests = len(self.submissions)
        n = 0

        for name in self.submissions:
            if callback:
                callback(name, n, numTests)

            self.submissions[name].runTest(test)
            n += 1

        if callback:
            callback(name, n, numTests)
        return n

    def runTestsOnSub(self, sub, callback=None, numTests=None, n=0):
        """Run the whole rubric on this submission.

        @param sub The submission to run tests on 
        """
        assert(isinstance(sub, submissions.Submission))
        if numTests is None:
            numTests = len(self.rubric.getTopLevel())

        for test in self.rubric.topLevel:
            if callback:
                callback(sub.getUsername(), n, numTests)

            sub.runTest(test)
            n += 1
                
        if callback:
            callback(sub.getUsername(), n, numTests)
            
        return n

    def runAllTests(self, callback=None):
        """Run the whole rubric on all submissions.
        """
        numTests = len(self.rubric.getTopLevel()) * len(self.submissions)

        n = 0
        for name in self.submissions:
            n = self.runTestsOnSub(self.submissions[name], callback,
                                   numTests, n)
        return n

    def genMailbacks(self, projName=None, callback=None, numOps=None):
        """Generate mailbacks for all students"""

        if projName is None:
            projName = self.assignmentName

        if numOps == None:
            numOps = len(self.submissions)

        n = 0
        names = self.submissions.keys()
        names.sort()
        for name in names:
            s = self.submissions[name]
            if callback:
                callback("Writing " + s.getUsername(), n, numOps)
                
            s.genMailbacks(projName, self.rubric, self.agarPath)

            n += 1

        if callback:
            callback("Writing " + s.getUsername(), n, numOps)

        return n

    def sendMailbacks(self, projName=None, callback=None):
        numOps = 2 * len(self.submissions)                         

        if projName is None:
            projName = self.assignmentName

        n = self.genMailbacks(projName, callback, numOps)

        names = self.submissions.keys()
        names.sort()
        for name in names:
            s = self.submissions[name]
            if callback:
                callback("Sending " + s.getUsername(), n, numOps)

            s.sendMailbacks(projName, self.reply, self.cclist, self.domain)
            n += 1

            # Don't overwhelm the mail server
            time.sleep(.5)

        if callback:
            callback("Sending " + s.getUsername(), n, numOps)
        
def defaultIsOntime(x):
    """Returns true iff ONTIME is a directory in x

    >>> defaultIsOntime("test1/student1")
    True
    >>> defaultIsOntime("test1/student2")
    False
    """
    return os.path.isdir(x + "/ONTIME")

def defaultGetFiles(x):
    """Returns the files in LATE/ if it exists, or ONTIME otherwise.
    Assumes that at least one of those exist, will throw an exception
    otherwise.  All files will be prefixed with the directory provided.

    >>> defaultGetFiles("test1/student1")
    ['test1/student1/ONTIME/Makefile', 'test1/student1/ONTIME/s1.cc', 'test1/student1/ONTIME/s1.h']
    >>> defaultGetFiles("test1/student2")
    ['test1/student2/LATE/s2.cc']
    >>> defaultGetFiles("test1/student3")
    ['test1/student3/LATE/s3.h']
    """

    x = os.path.normpath(x)
    
    if os.path.isdir(x + "/LATE") or os.path.islink(x + "/LATE"):
        ret = [x + "/LATE/" + i for i in os.listdir(x + "/LATE")]
        ret.sort()
        return [x for x in ret if os.path.isfile(x) and not\
                x.endswith("graderExe")]
    
    ret = [x + "/ONTIME/" + i for i in os.listdir(x + "/ONTIME")]
    ret.sort()
    return [x for x in ret if os.path.isfile(x) and \
            not x.endswith("graderExe") and \
            not x.endswith("grader.pdf") and \
            not x.endswith(".ppm")]

            
def defaultIsSubDir(x):
    """ Returns true if this isn't a hidden directory and if it has
    either"LATE" or "ONTIME" as a subdirectory/link to a subdirectory.

    >>> defaultIsSubDir("test1/student1")
    True
    >>> defaultIsSubDir("test1/bad1")
    False
    >>> defaultIsSubDir("test1/student2")
    True
    >>> defaultIsSubDir("test1/student3")
    True
    >>> defaultIsSubDir("test1/bad2")
    False
    >>> defaultIsSubDir("test1/bad3")
    False
    """
    
    return (x[x.index("/") + 1] != "." and os.path.isdir(x) and \
            (("ONTIME" in os.listdir(x) and (os.path.isdir(x + "/ONTIME") or \
                                             os.path.islink(x + "/ONTIME")))or\
             ("LATE"   in os.listdir(x) and (os.path.isdir(x + "/LATE") or \
                                             os.path.islink(x + "/LATE")))))

if __name__ == "__main__":
    import doctest
    doctest.testmod()
    failures, tests = doctest.testfile("tests.txt")

    if failures:
        sys.exit(1)

    print "No failures, checking coverage"
    import pycoverfuncs
    pycoverfuncs.cover(".coverage")
    for line in open("tests.txt"):
        if line.startswith(">>>"):
            exec(line[4:])

    commands.getstatusoutput("python pyviewc.py -i .coverage")

    for fn in os.listdir("."):
        if fn.endswith(".cov"):
            if not os.access(fn[:-4], os.R_OK):
                os.unlink(fn)
                continue
            
            commands.getoutput("mv %s coverage/%s" % (fn, fn))

            lines = 0
            good = 0
            for line in open("coverage/" + fn):
                if not line.strip() or line.strip()[0] == "#" or \
                   line.strip().startswith('"""'):
                    continue                 

                lines += 1
                if line[0] in ">":
                    good += 1
                    
            print fn, "%d/%d (%.2f%%)" % (good, lines, float(good)/lines)
