from xml.sax.saxutils import escape, quoteattr

import tests
import rubric

class UserComments:
    """A dictionary wrapper for Comment objects.
    """

    def __init__(self):
        self.comments = {}

    def __iter__(self):
        return iter(self.comments)

    def __getitem__(self, key):
        return self.comments[key]

    def __len__(self):
        return len(self.comments)

    def __setitem__(self, key, val):
        self.comments[key] = val

    def __delitem__(self, key):
        del self.comments[key]

    def keys(self):
        return self.comments.keys()

    def set(self, newComments, rubric):
        self.comments = newComments

        # Fix up which tests are connected where
        for c in self.comments.values():
            c.test.rubric = rubric

    def merge(self, newComments, preferOld, subs):
        for cname in newComments:
            if cname in self.comments:
                if preferOld:
                    for name in subs:
                        sub = subs[name]
                        toRemove = []
                        toAppend = []
                        for c in sub.comments:
                            if c == self.comments[cname]:
                                toRemove.append(c)
                                toAppend.append(newComments[cname])

                        for r in toRemove:
                            sub.comments.removeComment(r)
                        for c in toAppend:
                            sub.comments.addComment(c)
                            
                    self.comments[cname] = newComments[cname]
            else:
                self.comments[cname] = newComments[cname]

    def toXML(self):
        ret = []
        ret.append("<user_comments>")
        for c in self.comments:
            ret.append('<user_comment name=%s>' % quoteattr(c))
            ret += self.comments[c].toXML()
            ret.append('</user_comment>')
        ret.append('</user_comments>')
        return ret

class Comments:
    """A read-only container for Comment objects.

    Members:
    self.sub -- The submission we are attached to
    self.comments -- The list of comments we have
    """
    def __init__(self, sub):
        self.sub = sub
        self.comments = []
        self.locations = []

    def __iter__(self):
        return iter(self.comments)

    def __getitem__(self, ind):
        return self.comments[ind]

    def __len__(self):
        return len(self.comments)

    def index(self, c):
        return self.comments.index(c)

    def removeAutoComments(self, test):
        """Remove any comment that was automatically assigned, and was
        related to the given test."""        
        for comment in self.comments[:]:
            if comment.test == test:
                self.comments.remove(comment)

    def clear(self):
        """Drop all comments from this container."""
        self.comments = []

    def removeComment(self, comment):
        """Remove this one from the list"""
        if comment in self.comments:
            self.comments.remove(comment)
        
    def addComment(self, comment):
        """Add this to the list of comments"""
        assert(isinstance(comment, Comment))
        self.comments.append(comment)

    def addCommentLoc(self, comment, x, y, page):
        if len(self.comments) > len(self.locations):
            self.locations += [None] * (len(self.comments) - \
                                        len(self.locations))

        assert(len(self.comments) == len(self.locations))
        self.addComment(comment)

        self.locations.append((x, y, page))

    def hasLoc(self, ind):
        if ind >= len(self.locations) or self.locations[ind] == None:
            return False
        return True

    def setLoc(self, ind, x, y, page):
        while ind >= len(self.locations):
            self.locations += [None]
        self.locations[ind] = (x, y, page)

    def getLoc(self, c):
        ind = self.index(c)

        return self.locations[ind]

    def getCommentLocs(self):
        return [[self.comments[ind], self.locations[ind]] \
                for ind in range(len(self.comments)) \
                if ind < len(self.locations) and \
                self.locations[ind] is not None]

    def hasAnnotations(self):
        for x in self.locations:
            if x != None:
                return True
        return False            

    def updateTestScores(self, test, ratio):
        """The max score for the given test has changed, update all
        auto-assigned additive comments by the change amount.
        """
        for comment in self.comments:
            if comment.test == test and comment.auto and comment.op == "+":
                comment.value *= ratio                

    def hasTestComments(self, test):
        """Do we have any comments pertaining to this test?"""
        for comment in self.comments:
            if comment.test == test:
                return True
        return False

    def commentTestText(self, test):
        """Return text descriptions of the comments pertaining to the
        given test as a string, suitable for mailbacks."""

        ret = ""
        for comment in self.comments:
            if comment.test == test:
                ret += comment.getMailbackText() + "\n \n"

        return ret[:-2]
        
    def getScore(self, test=None):
        """Get the total score for these comments."""
        if test == None:
            # find the rubric: it's the .rubric member of a test
            tests = list(set([comment.test for comment in self.comments]))
            tests = [x for x in tests if x.rubric.isTopLevel(x)]
            if not tests:
                return 0.0
            rubric = tests[0].rubric
            assert(rubric != None)
            tests = rubric.getTopLevel()
            return float(sum([self.getScore(test) for test in tests]))
        
        total = 0.0
        mult = 1.0
        ret = None

        for comment in self.comments:
            if test.hasChild(comment.test) or comment.appliesAll:
                if comment.op == "+":
                    total += comment.value
                elif comment.op == "*":
                    mult *= comment.value
                elif comment.op == "=":
                    # Can only allow a single "set value" comment at a time,
                    # otherwise the results are undefined.
                    if ret != None:
                        raise RuntimeError
                    ret = comment.value
                else:
                    raise ValueError

        if ret != None:
            return float(ret)
        
        return float(total * mult)

    def getScores(self, rubric):
        """Return the scores for this rubric."""
        return [self.getScore(test) for test in rubric.getTopLevel()]
    
class Comment:
    """The Comment class is effectively a struct storing the values
    relating to this comment.

    Members:
    test -- which test we are related to
    auto -- were we automatically assigned?
    value -- the value of the comment
    comment -- the text of the comment    
    op -- "+", "=", or "*", determining whether this is additive, absolute,
    or multiplicative

    """
    def __init__(self, test, auto, value, comment, op, appliesToAll=False):
        assert(isinstance(test, rubric.RubricTest))
        assert(auto in [True, False])
        assert(type(value) in [type(0), type(0.0)])
        assert(type(comment) == type(""))
        assert(op in ["+", "*", "="])
        self.test = test
        self.auto = auto
        self.value = value
        self.comment = comment
        self.op = op

        self.appliesAll = appliesToAll

    def toXML(self, location=None):
        ret = []
        
        if self.auto:
            auto = 1
        else:
            auto = 0
            
        if self.appliesAll:
            all = 1
        else:
            all = 0

        if location == None:
            ret.append('<comment testname=%s auto="%d" value="%f" op="%s" all="%d">' % (quoteattr(self.test.getName()), auto, self.value, self.op, all))
        else:
            ret.append('<comment testname=%s auto="%d" value="%f" op="%s" all="%d" x="%d" y="%d" page="%d">' % (quoteattr(self.test.getName()), auto, self.value, self.op, all, location[0], location[1], location[2]))
            
        ret += [escape(x) for x in self.comment.split("\n")]
        ret.append('</comment>')
        return ret

    def __eq__(self, other):
        return self.test == other.test and \
               self.auto == other.auto and \
               self.value == other.value and \
               self.comment == other.comment and \
               self.op == other.op and \
               self.appliesAll == other.appliesAll

    def passed(self):
        """Is this a "success" or not?"""
        return self.value > 0
        
    def getMailbackText(self):
        """Get the mailback text for this comment."""
        if self.op == "+":
            return self.comment + "\nPoints: " + str(self.value)
        elif self.op == "*":
            if self.appliesAll:
                return self.comment + "\nPercent: " + str(self.value * 100) +\
                       " (Applies to all rubric items)"
            else:
                return self.comment + "\nPercent: " + str(self.value * 100) +\
                       " (This rubric item only)"                
        elif self.op == "=":
            if self.appliesAll:
                return self.comment + "\nSet Score: " + str(self.value) +\
                       " (Applies to all rubric items)"
            else:
                return self.comment + "\nSet Score: " + str(self.value) +\
                       " (This rubric item only)"                

        assert(False)
        
