import sys
import gzip
import re
import string

import submissions
import comments
import rubric

from xml.parsers import expat

class Element:
    'A parsed XML element'
    def __init__(self,name,attributes):
        'Element constructor'
        # The element's tag name
        self.name = name
        # The element's attribute dictionary
        self.attributes = attributes
        # The element's cdata
        self.cdata = ''
        # The element's child element list (sequence)
        self.children = []

    def __iter__(self):
        return iter(self.children)

    def __getitem__(self, v):
        if type(v) == type(0):
            return self.children[v]
        return [x for x in self.children if x.name == v]
        
    def AddChild(self,element):
        'Add a reference to a child element'
        self.children.append(element)
        
    def getAttribute(self,key):
        'Get an attribute value'
        return str(self.attributes.get(key))
    
    def getData(self):
        'Get the cdata'
        return str(self.cdata)
        
    def getElements(self,name=''):
        'Get a list of child elements'
        #If no tag name is specified, return the all children
        if not name:
            return self.children
        else:
            # else return only those children with a matching tag name
            elements = []
            for element in self.children:
                if element.name == name:
                    elements.append(element)
            return elements

class Xml2Obj:
    'XML to Object'
    def __init__(self):
        self.root = None
        self.nodeStack = []
        
    def StartElement(self,name,attributes):
        'SAX start element even handler'
        # Instantiate an Element object
        element = Element(name.encode(),attributes)
        
        # Push element onto the stack and make it a child of parent
        if len(self.nodeStack) > 0:
            parent = self.nodeStack[-1]
            parent.AddChild(element)
        else:
            self.root = element
        self.nodeStack.append(element)
        
    def EndElement(self,name):
        'SAX end element event handler'
        self.nodeStack = self.nodeStack[:-1]

    def CharacterData(self,data):
        'SAX character data event handler'
        if string.strip(data):
            data = data.encode()
            element = self.nodeStack[-1]
            element.cdata += data
            return

    def Parse(self,data):
        # Create a SAX parser
        Parser = expat.ParserCreate()

        # SAX event handlers
        Parser.StartElementHandler = self.StartElement
        Parser.EndElementHandler = self.EndElement
        Parser.CharacterDataHandler = self.CharacterData

        # Parse the XML File
        ParserStatus = Parser.Parse(data, 1)
        
        return self.root
    
def Load(xml):
    xml = " ".join(xml)
    parser = Xml2Obj()
    
    itemTree = parser.Parse(xml)

    version = itemTree["version"][0].getAttribute("full")

    rubricData, order = parseRubric(itemTree["rubric"][0])
    cs = parseComments(itemTree["user_comments"][0], rubricData)
    subs = parseSubs(itemTree["submissions"][0], cs, rubricData)

    basedir = itemTree["project"][0].getAttribute("basedir")

    try:
        reply = itemTree["mail"][0].getAttribute("reply")
    except:
        reply = ""

    try:
        cclist = itemTree["mail"][0].getAttribute("cclist")
    except:
        cclist = ""

    try:
        domain = itemTree["mail"][0].getAttribute("domain")
    except:
        domain = "cs.ucr.edu"

    try:
        name = itemTree["project"][0].getAttribute("name")
    except:
        name = "Agar Project"

    
    return version, subs, cs, rubricData, order, \
           reply, cclist, domain, name, basedir

def parseRubric(rData):
    # Return a dictionary of RubricItem objects, the Rubric will
    # handle extracting the toplevel info itself.
    ret = {}
    order = []

    for ri in rData["rubricTests"][0]["rubricItem"]:
        # It is a bit icky to instantiate these without the rubric or
        # tests, but the actual rubric object will fix it all up when
        # these RIs are passed to it.
        newRI = rubric.RubricTest(None, ri.getAttribute("testName"),
                                  float(ri.getAttribute("points")),
                                  ri.getAttribute("name"))
        newRI.fileMask = ri.getAttribute("fileMask")

        for child in ri["children"][0]["child"]:
            newRI.children.append(child.getData())

        for cs in ri["autocomments"]:
            for c in cs["comment"]:
                newComment = \
                           comments.Comment(newRI,
                                            (c.getAttribute("auto")== "1"),
                                            float(c.getAttribute("value")),
                                            c.getData(),
                                            c.getAttribute("op"),
                                            (c.getAttribute("all") == "1"))

                if cs.getAttribute("success") == "1":
                    newRI.successComments.append(newComment)
                else:
                    newRI.failureComments.append(newComment)

        newRI.config = {}
        for cfg in ri["config"][0]["configitem"]:
            newRI.config[cfg.getAttribute("key")] = cfg.getAttribute("value")
            
        ret[newRI.name] = newRI

    for tn in rData["topLevel"][0]["testname"]:
        order.append(str(tn.getData()))

    return ret, order

def parseComments(cData, rubric):
    # Return a dictionary of Comment objects
    ret = {}

    for uc in cData["user_comment"]:
        # Name is stored in the user_comment wrapper
        name = uc.getAttribute("name")

        # Get the comment itself
        c = uc["comment"][0]

        # Build the new comment
        newComment = comments.Comment(rubric[c.getAttribute("testname")],
                                      (c.getAttribute("auto") == "1"),
                                      float(c.getAttribute("value")),
                                      c.getData(),
                                      c.getAttribute("op"),
                                      (c.getAttribute("all") == "1"))
        
        ret[name] = newComment

    return ret

def parseSubs(subs, cs, rts):
    # Return a dictionary of Submission objects
    ret = {}

    for s in subs["submission"]:
        files = []
        for fn in s["files"][0]["filename"]:
            files.append(fn.getData())

        newSub = submissions.Submission(s.getAttribute("name"),
                                        files,
                                        s.getAttribute("basedir"))

        newSub.executable = s.getAttribute("executable")
        if newSub.executable == "None":
            newSub.executable = None

        try:
            pdf = s["annotation_info"][0].getAttribute("pdf")
            numPages = int(s["annotation_info"][0].getAttribute("numpages"))
            newSub.pdf = pdf
            newSub.numPages = numPages
        except:
            pass

        usernames = []
        for u in s["usernames"][0]["username"]:
            usernames.append(u.getData())

        for i in usernames:
            newSub.addUsername(i)

        for c in s["comments"][0]["comment"]:
            newComment = comments.Comment(rts[c.getAttribute("testname")],
                                          (c.getAttribute("auto") == "1"),
                                          float(c.getAttribute("value")),
                                          c.getData(),
                                          c.getAttribute("op"),
                                          (c.getAttribute("all") == "1"))

            hasLoc = False
            if not newComment.auto:
                found = False
                for i in cs.values():
                    if i == newComment:
                        newComment = i
                        found = True
                        break
                assert(found)

                try:
                    x = int(c.getAttribute("x"))
                    y = int(c.getAttribute("y"))
                    page = int(c.getAttribute("page"))
                    hasLoc = True
                except KeyError:
                    pass
                except ValueError:
                    pass

            if hasLoc:
                newSub.comments.addCommentLoc(newComment, x, y, page)
            else:
                newSub.comments.addComment(newComment)
            
        ret[newSub.getUsername()] = newSub


    return ret
