import wx

class TreeNode:
    """Basically an n-ary tree with a few data members: name (what is
    displayed), id (the TreeId of the node), expanded (boolean), tag
    (whether we've seen this since the last time we untagged, used for
    pruning), and data (associated data object)"""
    def __init__(self, name, id,
                 expanded=False, children=None,
                 tag=True, data=None):
        self.name = name
        self.id = id
        self.expanded = expanded

        if children is None:
            self.children = []
        else:
            self.children = []

        self.tag = tag
        self.data = data

class MemTree(wx.TreeCtrl):
    """A tree class that remembers (usually) when a node has been
    expanded, and handles some statefulness of the tree."""
    def __init__(self, rootName, *args, **kwds):
        wx.TreeCtrl.__init__(self, *args, **kwds)

        rootID = self.AddRoot(rootName)
        self.tree = TreeNode(rootName, rootID, True)

        wx.EVT_TREE_ITEM_COLLAPSED(self, -1, self.onCollapse)
        wx.EVT_TREE_ITEM_EXPANDED(self, -1, self.onExpand)

        wx.EVT_TREE_SEL_CHANGED(self, -1, self.onSelected)

        self.selectedPath = []

    def getDataById(self, id, cur=None):
        """Given a tree ID, find the associated data"""
        if cur == None:
            cur = self.tree

        if cur.id == id:
            return cur.data

        for child in cur.children:
            temp = self.getDataById(id, child)
            if temp is not None:
                return temp

        return None

    def onSelected(self, event):
        """Something was selected, so note the path of what was
        selected so we can keep that selected after a call to update"""
        cur = event.GetItem()
        rootId = self.tree.id

        self.selectedPath = []

        while cur != rootId:
            self.selectedPath = [str(self.GetItemText(cur))] + \
                                self.selectedPath
            cur = self.GetItemParent(cur)
        
    def unTag(self, cur=None):
        """Set all tags to false.  This will be followed by a call to
        update and a call to prune: anything that hasn't been tagged
        True by update will be pruned away"""
        if cur == None:
            cur = self.tree

        for child in cur.children:
            child.tag = False
            self.unTag(child)

        self.tree.tag = True

    def toggleExpanded(self, curId, val):
        """This node's expanded state is set to 'val'"""
        # Build the path to this one:
        path = [str(self.GetItemText(curId))]
        while curId != self.GetRootItem():
            curId = self.GetItemParent(curId)
            path = [str(self.GetItemText(curId))] + path

        # We don't need the root text
        path.pop(0)

        # It was the root
        if not path:
            self.tree.expanded = val

        # Now, with that path, traverse down the tree to toggle
        # expanded
        cur = self.tree
        while path:
            cur = self.nodeGetChild(cur, path[0])
            path.pop(0)

        cur.expanded = val

    def onCollapse(self, event):
        """Note that we've altered the expanded state"""
        self.toggleExpanded(event.GetItem(), False)

    def onExpand(self, event):
        """Note that we've altered the expanded state"""
        self.toggleExpanded(event.GetItem(), True)

    def pathAdd(self, path, expanded=False, data=None, ind=None):
        """Path is a list of strings.  Add all the missing things in
        that list, from root to leaf.  Sets tags for existing elements
        to True"""
        cur = self.tree

        for item in path:
            next = self.nodeGetChild(cur, item)
            if next is None:
                if ind is None or item != path[-1]:
                    newId = self.AppendItem(cur.id, item)

                    next = TreeNode(item, newId, expanded, data=data)
                    cur.children.append(next)
                else:
                    newId = self.InsertItemBefore(cur.id, ind, item)
                    next = TreeNode(item, newId, expanded, data=data)

                    cur.children.insert(ind, next)
            else:
                next.tag = True

            cur = next

    def prune(self, cur=None):
        """Remove everything that hasn't been seen recently (wasn't
        hit with a pathAdd)"""
        if cur is None:
            cur = self.tree

        for child in cur.children[:]:
            
            if not child.tag:
                cur.children.remove(child)
                continue
                
            self.prune(child)

    def nodeGetChild(self, cur, item):
        """Return the node with this name from the children of cur"""
        for node in cur.children:
            if node.name == item:
                return node
        return None

    def update(self):
        """Redraw the widget, keeping everything as stable as possible
        (selection, expanded nodes, etc)"""
        self.DeleteAllItems()

        self.tree.id = self.AddRoot(self.tree.name)
        
        for child in self.tree.children:
            self.drawTree(child, self.tree.id)

        self.Expand(self.tree.id)

        cur = self.tree
        selected = False
        for name in self.selectedPath:
            next = self.nodeGetChild(cur, name)

            if next is None:
                self.SelectItem(cur.id)
                selected = True
                break

            cur = next

        if not selected:
            self.SelectItem(cur.id)

    def drawTree(self, node, parentID):
        """Draw the widget"""
        node.id = self.AppendItem(parentID, node.name)

        for child in node.children:
            self.drawTree(child, node.id)
        
        if node.expanded:
            self.Expand(node.id)
    
