import glob
import os
import commands

from xml.sax.saxutils import escape, quoteattr

import comments
import rubric


class Submissions:
    """
    This is a container class for Submission objects.  It behaves
    essentially like a read-only dictionary.

    Major Members:
    self.submissions -- a dictionary mapping usernames to Submission objects

    Minor Members:
    self.basedir -- the directory containing all Submission objects
    self.config -- a reference to the parent config dictionary
    """
    
    def __init__(self, config, basedir, filemask=None):
        """
        Construct a new Submissions container.

        @param config the parent Agar object's self.config dictionary
        @param basedir the directory containing all the submissions
        """
        assert(type(basedir) == type("") and os.path.isdir(basedir))
        assert(type(config) == type({}))
        
        self.basedir = basedir
        self.config = config

        if filemask is None:
            self.fileMask = "*.c,*.cc,*.C,*.cpp,*.h,Makefile"
        else:
            self.fileMask = filemask

        # Find directories in basedir that match config["isSubmissionDir"]
        self.newSubs()

    def setFileMask(self, mask):
        self.fileMask = mask

    def getFileMask(self):
        return self.fileMask

    def set(self, newSubs, rubricData):
        self.submissions = newSubs

        # Fix up which tests are connected where
        for sub in self.submissions.values():
            for comment in sub.comments:
                assert(isinstance(comment.test, rubric.RubricTest))
                comment.test.rubric = rubricData

    def merge(self, newSubs):
        for name in newSubs:
            if name in self.submissions:
                for c in newSubs[name].comments:
                    self.submissions[name].addComment(c)
            else:
                self.submissions[name] = newSubs[name]

    def toXML(self):
        ret = []
        ret.append("<submissions>")
        names = self.submissions.keys()
        names.sort()

        for name in names:
            ret += self.submissions[name].toXML()

        ret.append("</submissions>")
        return ret

    def newSubs(self, basedir=None):
        """Set the submissions based on the given base directory"""
        if basedir == None:
            basedir = self.basedir
        
        subDirs = self.findSubs(basedir)

        # Setup the submissions
        self.submissions = {}
        getFiles = self.config["getFiles"]
        for s in subDirs:
            files = getFiles(basedir + "/" + s)
            goodFiles = []
            for filename in files:
                good = False
                for mask in self.fileMask.split(","):
                    mask = mask.strip()
                    if filename in glob.glob("%s/%s" % \
                                             (os.path.dirname(filename),mask)):
                        good = True
                        break
                if good:
                    goodFiles.append(filename)
                
            self.submissions[s] = Submission(s,
                                             goodFiles,
                                             basedir)

    def __iter__(self):
        return iter(self.submissions)

    def __getitem__(self, key):
        return self.submissions[key]

    def __len__(self):
        return len(self.submissions)

    def __delitem__(self, key):
        del self.submissions[key]

    def keys(self):
        return self.submissions.keys()

    def add(self, name):
        assert(name not in self.submissions)

        files = self.config["getFiles"](self.basedir + "/" + name)
        self.submissions[name] = Submission(name,
                                            files,
                                            self.basedir)

    def findSubs(self, basedir=None):
        """ Given a subdirectory (or defaulting to the cwd), find the
        submissions stored in that directory.

        Relies on: self.config["isSubmissionDir"]
        """        
        
        if basedir == None:
            basedir = self.basedir
            
        assert(type(basedir) == type(""))
        
        dirs = [x for x in os.listdir(basedir) \
                if os.path.isdir(basedir + "/" + x) and \
                self.config["isSubmissionDir"](basedir + "/" + x)]
        
        dirs.sort()
        return dirs

    def getUsernames(self):
        """Return a sorted list of the usernames"""
        ret = [x for x in self.submissions]
        ret.sort()
        return ret
        
class Submission:
    """A class representing a single submission.

    Major Members:
    self.comments -- the Comments container for this Submission
    """
    def __init__(self, username, files, basedir):
        """Construct a new Submission object.

        Parameters:
        username -- the name of the directory in basedir that this represents
        files -- the files in this directory that are part of the submission
        """

        assert(type(username) == type(""))
        assert(type(files) in [type([]), type(())])
        
        self.username = username
        self.files = files[:]
        self.comments = comments.Comments(self)

        self.executable = None

        self.basedir = basedir

        self.usernames = set()
        self.usernames.add(username)

        # These will be set during the conversion-for-annotation
        # process
        self.numPages = None
        self.pdf = None

    def toXML(self):
        ret = []
        ret.append('<submission name=%s executable=%s basedir=%s>' % \
                   (quoteattr(self.username),
                    quoteattr(str(self.executable)),
                    quoteattr(self.basedir)))

        if self.pdf is not None:
            ret.append('<annotation_info pdf=%s numpages="%d" />' % \
                       (quoteattr(self.pdf),
                        self.numPages))

        ret.append("<files>")
        for fn in self.files:
            ret.append("<filename>%s</filename>" % escape(fn))
        ret.append("</files>")

        ret.append("<comments>")
        for ind, comment in enumerate(self.comments):
            if self.comments.hasLoc(ind):
                ret += comment.toXML(self.comments.locations[ind])
            else:
                ret += comment.toXML()
        ret.append("</comments>")

        ret.append("<usernames>")
        for name in self.usernames:
            ret.append('<username>%s</username>' % escape(name))
        ret.append("</usernames>")
        ret.append("</submission>")

        return ret
    
    def getFiles(self):
        """Return a copy of the file list for this submission"""
        if self.executable:
            return self.files[:] + [self.executable]
        
        return self.files[:]

    def getPath(self):
        """Return the path of the directory storing the files in this
        submission."""
        return self.basedir + "/" + \
               os.path.dirname(os.path.commonprefix(self.files))

    def getUsername(self):
        """Return the primary username for this submission"""
        return self.username

    def getUsernames(self):
        """Return the whole list of usernames for this submission"""
        l = list(self.usernames)
        l.sort()
        return l

    def addUsername(self, newUsername):
        """Add to the username set for this submission."""
        self.usernames.add(newUsername)

    def removeUsername(self, name):
        """Remove a username from the username set for this
        submission.  Throws ValueError if the primary username is
        specified.  (The primary can't be removed, otherwise we
        wouldn't know where in basedir our files are located!)  Also
        throws ValueError if the given name isn't in the set.
        """
        if self.username == name or name not in self.usernames:
            raise ValueError

        self.usernames.remove(name)

    def sendMailbacks(self, projName, replyTo, cclist, domain):
        for name in self.usernames:
            self.sendMailback(projName, name, replyTo, cclist, domain)

    def sendMailback(self, projName, username, replyTo, cclist, domain):
        cmdLine = 'cat %s/%s.mailback | mutt -s "%s Results"' % \
                  (self.basedir, username, projName)

        if replyTo.count("@"):
            os.environ["REPLYTO"] = replyTo

        if cclist.strip():
            for cc in cclist.split(","):
                cc = cc.strip()
                cmdLine += " -c " + cc

        if self.comments.hasAnnotations():
            if os.access("%s/%s.annotated.pdf" % (self.basedir, username),
                         os.R_OK):
                cmdLine += " -a %s/%s.annotated.pdf" % (self.basedir, username)

        # TODO: Fix these for distribution
        username = "titus"
        cmdLine += " %s@%s" %  (username, "finn.winters-ca.net")

        print cmdLine
        commands.getoutput(cmdLine)

    def genMailbacks(self, projName, rubric, path):
        """Generate the mailbacks for this submission, one for each username"""
        for name in self.usernames:
            self.genMailback(projName, rubric, name, path)

    def genMailback(self, projName, rubric, username, path):
        """Generate the mailback for this username in the file
        'username.mailback'"""
        of = open("%s/%s.mailback" % (self.basedir, username), 'w')

        print >>of, self.getMailbackStr(projName, rubric, username)

        # TODO: Need to do the annotation of the submission here
        if self.pdf is not None:
            self.writeAnnotations(username, path)

        of.close()

    def getMailbackStr(self, projName, rubric, username):
        """Return the text of the mailback for this username"""
        lines = []
        lines.append("%s: Mailback for %s" % (projName, username))
        lines.append("Total Score: " + str(self.getScore()))
        lines.append("")

        for test in rubric.topLevel:
            lines.append("%s: " % test.getName() + str(self.getScore(test)))

        lines.append("")
        lines.append("Comments:")

        for test in rubric.topLevel:
            if self.comments.hasTestComments(test):
                lines.append("********* %s ********* " % test.getName())
                lines += self.comments.commentTestText(test).split("\n")
                lines.append("")
                    
        return "\n".join(lines)

    def convertForAnnotation(self, path):
        dirName = os.path.dirname(self.files[0])
        pdfFile = dirName + "/grader.pdf"
        cmdLine = "%s/all2pdf.py %s %s" % \
                  (path, pdfFile, " ".join(self.files))
        
        status, output = commands.getstatusoutput(cmdLine)
        if status:
            print "Error converting to PDF", output
            self.pdf = None
            self.setNumPages(None)
            return
            
        self.pdf = pdfFile
        
        cmdLine = "%s/convertPDF.sh %s %s" % \
                  (path, os.path.dirname(self.pdf),
                   os.path.dirname(self.pdf) + "/grader.pdf")

        pages = commands.getoutput(cmdLine)

        try:
            self.numPages = int(pages)
        except ValueError:
            print "Error converting to PPM", pages
            self.pdf = None
            self.numPages = None            

    def writeAnnotations(self, name, path):
        filename = os.path.dirname(self.pdf) + "/" + \
                   self.getUsername() + "-1.ppm"
        fh = open(filename)
        fh.readline()
        width, height = [int(x) for x in fh.readline().split()]
        fh.close()

        out = open(self.basedir + "/" + name + ".annotate.xml", 'w')        
                
        print >>out, """<?xml version="1.0" encoding="ISO-8859-1"?>
<annotationSpec xmlns="http://xml.cs.ucr.edu/annotate"
xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
xsi:schemaLocation="http://xml.cs.ucr.edu/annotate/annotationSpec-0.1.xsd">

<version>0.1</version>
<annotationSpace>
  <width>%d</width>
  <height>%d</height>
</annotationSpace>

<annotationList>
""" % (width, height)

        for comment, (cx, cy, page) in self.comments.getCommentLocs():
            cx += 30
            cy += 40
            cx = min(cx, 600)
            cy = min(cy, 800)
            if comment.value < 0:
                r, g, b = (200, 0, 0)
            elif comment.value == 0:
                r, g, b = (200, 200, 200)
            else:
                r, g, b = (0, 200, 0)
                
            print >>out, """<annotation>
    <note>
      <pageNum>%d</pageNum>
      <center>
        <xCoord>%d</xCoord>
        <yCoord>%d</yCoord>
      </center>
      <content>
        %s
      </content>
      <rgbColor>
        <rgbRed>%d</rgbRed>
        <rgbGreen>%d</rgbGreen>
        <rgbBlue>%d</rgbBlue>
      </rgbColor>
    </note>
  </annotation>
  """ % (page, cx, cy, comment.comment, r, g, b)

        print >>out, "  </annotationList>\n</annotationSpec>\n\n"
        out.close()

        tempPS = os.path.dirname(self.pdf) + "/" + \
                 name + ".ps"
        annotatedPS = tempPS[:-2] + "annotated.ps"
        annotatedPDF = os.path.dirname(os.path.realpath(out.name)) + "/" + \
                       name + ".annotated.pdf"
        (status, output) = commands.getstatusoutput("pdf2ps %s %s" % \
                                                    (self.pdf,
                                                     tempPS))
        if status:
            print "Failure to convert to temp PS file:"
            print output
            return

        annotate = path + "/annotate"
        (status, output) = commands.getstatusoutput("%s %s %s %s" % \
                                                    (annotate,
                                                     tempPS,
                                                     out.name,
                                                     annotatedPS))
        
        if status:
            print "Failure to annotate PS file:"
            print output
            return

        (status, output) = commands.getstatusoutput("ps2pdf %s %s" % \
                                                    (annotatedPS,
                                                     annotatedPDF))
        
        os.unlink(tempPS)
        os.unlink(annotatedPS)
        os.unlink(out.name)

        return name + ": Converted PDF successfully"        

    def removeAutoComments(self, test):
        """Remove any auto-assigned comments that came from this test"""
        self.comments.removeAutoComments(test)        

    def removeComment(self, comment):
        """If we've got a copy of this commment, remove it."""
        self.comments.removeComment(comment)
    
    def runTest(self, test):
        """Run the given test, removing any old auto-generated
        comments in the process."""

        # Make sure we're clean
        self.removeAutoComments(test)

        # Run the test
        result = test.runTest(self)

        # Based on the results, store the appropriate comment(s)
        if type(result) == type([]):
            for newComment in result:
                self.comments.addComment(newComment)
        else:
            assert(isinstance(result, comments.Comment))
            self.comments.addComment(result)

    def getScore(self, test=None):
        """Return the score for this submission"""
        return self.comments.getScore(test)

    def getScores(self, rubric, username=None):
        """Return a detailed list of scores: username, total, and a
        score for each toplevel test."""
        if username == None:
            username = self.getUsername()
            
        return [username, self.getScore()] + \
               self.comments.getScores(rubric)

    def updateTestScores(self, test, ratio):
        """The max value for the given test has changed, update any
        test results associated with that test.
        """
        self.comments.updateTestScores(test, ratio)

    def getPath(self):
        return os.path.dirname(self.files[0])
