import os
import string
import commands


# Wrapping the dictionary of tests so that it is read-only
class Tests:
    """A read-only container class for Test objects.

    Major members:
    self.tests: a dictionary mapping from test names to Test objects
    """
    def __init__(self, dirName):
        self.tests = getTests(dirName)

    def __iter__(self):
        return iter(self.tests)

    def __getitem__(self, key):
        return self.tests[key]

    def __len__(self):
        return len(self.tests)

class Test:
    """A single Test object corresponds to an executable in the tests/
    directory in an Agar installation.

    Members:
    * self.fileName -- the name of the executable
    * self.name -- the name reported by the test
    * self.pick -- either min or max, as chosen by the test    
    """
    def __init__(self, fileName):
        self.fileName = fileName
        output = commands.getoutput(fileName + " --defaults").split("\n")

        # Tests pick their own name (line 1)
        self.name = output.pop(0).strip()

        # Tests also set their default behavior for choosing the comment
        # when multiple files match the filemask.
        if output.pop(0).strip() == "one":
            self.allAtOnce = False
        else:
            self.allAtOnce = True

        # Tests have a default filemask (line 3)
        self.fileMask = output.pop(0).strip()

        # Get the default configuration for this test
        try:
            self.defaultConfig = eval(output[0].strip())
            assert(type(self.defaultConfig) == type({}))
            output.pop(0)
            for key in self.defaultConfig:
                self.defaultConfig[key] = str(self.defaultConfig[key])
        except:
            self.defaultConfig = {}
        
        # Parse remaining usage message and deal with parameters
        good = [x for x in output if "--" in x]
        
        self.singletons = []
        self.attributes = []
        self.files = []

        for line in good:
            if "=" in line:
                s = ""
                for char in line[2:]:
                    if char == "-":
                        s += " "
                    elif char in ",.[]":
                        pass
                    elif char == "=":
                        ind = line.index("=") + 1
                        if line[ind:ind + 4].upper() == "FILE":
                            self.files.append(s)
                        else:
                            self.attributes.append(s)
                        break
                    else:
                        s += char
            else:
                s = ""
                for char in line.strip()[2:]:
                    if char == "-":
                        s += " "
                    elif char == "." or char == ",":
                        pass
                    elif char in string.whitespace:
                        break
                    else:
                        s += char

                self.singletons.append(s)

        self.singletons.sort()
        self.attributes.sort()
        self.files.sort()

    def run(self, params, submission):
        """Run this test with the given parameters."""
        inQuotes = False
        params = list(params)
        for c in range(len(params)):
            if params[c] == "'":
                inQuotes = not inQuotes
            if inQuotes and params[c] == '_':
                params[c] = ' '
        params = "".join(params)
                
        
        #if self.fileName.endswith("regexp.py"):
        #    print self.fileName + " " + params
        return commands.getstatusoutput(self.fileName + " " + params)

    def getFlagParams(self):
        """Return the name of the parameters that are just flags."""
        return self.singletons

    def getFileParams(self):
        """Return the list of parameters that take a file as an argument."""
        return self.files

    def getValParams(self):
        """Return the list of value paramters that do NOT take a file."""
        return self.attributes

    def getParams(self):
        """Return the name of all the parameters for this test"""
        temp = self.getFlagParams() + \
               self.getFileParams() + \
               self.getValParams()
        temp.sort()
        return temp

    def getName(self):
        """What is the name of this Test?"""
        return self.name

class CompileTest(Test):
    """A special Test object to do compilation.
    """
    def __init__(self):
        # Set the name
        self.name = "Compile"

        # Tests have a default filemask (line 3)
        self.fileMask = "*.C,*.cc,*.cpp"
        self.allAtOnce = True

        self.defaultConfig = {'Compiler': 'g++',
                              'Compiler Flags': '-W -Wall -Werror -pedantic',
                              'Output Flag': '-o'}

    def run(self, params, submission):
        """Compile, given the appropriate flags.  Updates the
        submission.executable member if successful."""
        # In params we are going to get parameters for things like
        # --Compiler= and --Compiler Flags=, so we're gonna fix those
        # up into a proper command line.
        tokens = params.split()

        submission.executable = None
        
        compiler = None
        outputFlag = ""
        flags = ""

        # This is to catch the files
        otherTokens = []

        # The point value is the last token
        points = float(tokens[-1])
        tokens.pop()

        for token in tokens:
            if "=" in token:
                var, val = token.split("=")
                val = val.replace("'", "")
                val = val.replace("_", " ")
                
                if var == "--Compiler":
                    compiler = val
                elif var == "--Compiler-Flags":
                    flags = val
                elif var == "--Output-Flag":
                    outputFlag = val
                else:
                    otherTokens.append(var + "=" + val)
            else:
                otherTokens.append(token)
                
        assert(type(compiler) == type(""))

        studentPath = os.path.dirname(submission.getFiles()[0])
        cmdLine = "%s %s %s/graderExe %s %s" % (compiler, outputFlag,
                                                studentPath, flags,
                                                " ".join(otherTokens))

        s, o = commands.getstatusoutput(cmdLine)
        if s:
            return s, "0\n" + o
        else:
            submission.executable = studentPath + "/graderExe"
            return 0, str(points) + "\nCompiled Successfully"
        

    def getFlagParams(self):
        """No flag parameters for compilation."""
        return []

    def getFileParams(self):
        """We need to know which compiler to use."""
        return ["Compiler"]

    def getValParams(self):
        """We also need to have compiler flags available, and which
        flag specifies the output file name."""
        return ["Compiler Flags", "Output Flag"]

    def getName(self):
        """What is the name of this Test?"""
        return "Compile"

class MakeTest(Test):
    """A special Test object to do compilation by invoking make(1)
    """
    def __init__(self):
        # Set the name
        self.name = "Invoke Make"

        # Tests have a default filemask (line 3)
        self.fileMask = "*"
        self.allAtOnce = True

        self.defaultConfig = {'Delete Binary Executables': True,
                              'Make Flags': ""}

    def run(self, params, submission):
        """Compile, given the appropriate flags.  Updates the
        submission.executable member if successful."""
        # In params we are going to get parameters for things like
        # --Compiler= and --Compiler Flags=, so we're gonna fix those
        # up into a proper command line.
        submission.executable = None

        curDir = submission.getPath()

        if "--Delete-Binary-Executables" in params:
            ind = params.index("--Delete-Binary-Executables")
            params = params[:ind] + \
                     params[ind + len("--Delete-Binary-Executables") + 1:]

            for filename in os.listdir(curDir):
                if filename == "graderExe" or \
                       os.path.isdir(curDir + "/" + filename):
                    continue
                
                output = commands.getoutput("file " + curDir + "/" + filename)
                if "ELF" in output and "executable" in output:
                    os.unlink(curDir + "/" + filename)

        points = float(params.split()[-1])

        if "--Make-Flags=" in params:
            params = params[:ind] + \
                     params[ind + len("--Make-Flags=") + 1:params.index("'", ind + len("--Make-Flags=") + 1)]
        else:
            params = ""

        oldFiles = os.listdir(curDir)
                    
        cmdLine = "cd %s ; make %s" % (curDir, params)

        s, o = commands.getstatusoutput(cmdLine)
        if s:
            return s, "0\n" + o
        else:
            files = [x for x in os.listdir(curDir) if x != "graderExe" and \
                     x not in oldFiles and \
                     not os.path.isdir(curDir + "/" + x) and \
                     os.access(curDir + "/" + x, os.X_OK)]

            exes = []
            for filename in files:
                output = commands.getoutput("file " + curDir + "/" + filename)
                if "ELF" in output and "executable" in output:
                    exes.append(filename)

            if len(exes) == 1:
                try:
                    os.unlink(curDir + "/graderExe")
                except:
                    pass

                os.link(curDir + "/" + exes[0], curDir + "/graderExe")
                os.unlink(curDir + "/" + exes[0])
                                
                submission.executable = curDir + "/graderExe"
                return 0, str(points) + "\nCompiled Successfully"
            else:
                retStr = "0\nFound %d executables" % len(exes) + \
                         "after calling make, aborting."
                return 1, retStr

    def getFlagParams(self):
        """No flag parameters for compilation."""
        return ["Delete Binary Executables"]

    def getFileParams(self):
        """We need to know which compiler to use."""
        return []

    def getValParams(self):
        """We also need to have compiler flags available, and which
        flag specifies the output file name."""
        return ["Make Flags"]

    def getName(self):
        """What is the name of this Test?"""
        return "Invoke Make"


class SuccessTest(Test):
    """A special Test class for 'always succeeds'.  Mostly used for
    situations where the points are assigned by default and then taken
    away by user comments (like grading written work).

    Members:
    * self.name -- the name reported by the test
    """
    def __init__(self):
        # Tests pick their own name (line 1)
        self.name = "Success"

        # Tests have a default filemask (line 3)
        self.fileMask = "*"

        # Might as well do all at once, we know the result
        self.allAtOnce = True

        self.defaultConfig = {}

    def run(self, params, submission):
        """Return success."""
        return (0, params.split()[-1] + "\nDefault points")

    def getFlagParams(self):
        """Return empty, we take no parameters."""
        return []

    def getFileParams(self):
        """Return empty, we take no parameters."""
        return []

    def getValParams(self):
        """Return empty, we take no parameters."""
        return []

class FailureTest(Test):
    """A special Test class for 'always fails'.  Mostly used for
    situations where a collection of tests should be run but not
    reported as top-level.

    Members:
    * self.name -- the name reported by the test
    """
    def __init__(self):
        # Tests pick their own name (line 1)
        self.name = "Failure"

        # Tests have a default filemask (line 3)
        self.fileMask = "*"

        # Might as well do all at once, we know the result
        self.allAtOnce = True

        self.defaultConfig = {}

    def run(self, params, submission):
        "Return failure."""
        return (0, "0\nNo default points")

    def getFlagParams(self):
        """Return empty, we take no parameters."""
        return []

    def getFileParams(self):
        """Return empty, we take no parameters."""
        return []

    def getValParams(self):
        """Return empty, we take no parameters."""
        return []
    
def isTest(fileName):
    """Given a filename, see if we can run it and parse the results
    (that is, it is an Agar test)."""
    # Don't run any backups
    if fileName[-1] == "~":
        return False

    if os.path.isdir(fileName):
        return False
    
    (status, output) = commands.getstatusoutput(fileName + " --defaults")
    return status == 2 * 256 and output != ""

def getTests(dirName):
    """Find every executable in this directory that behaves like an
    Agar test.  Also include all of the specialized tests (Success,
    Failure, etc).

    Return a dictionary mapping test names (as identified by the
    tests) to Test objects."""
    tests = [Test(dirName + "/" + x) for x in os.listdir(dirName) \
             if isTest(dirName + "/" + x)]

    tests += [SuccessTest(), FailureTest(), CompileTest(), MakeTest()]

    ret = {}
    for test in tests:
        ret[test.getName()] = test

    return ret
